#include "bin2abc.h"
#include "asmsrc/bac80.h"
#include "asmsrc/bac80r.h"

struct basic_block {
    FILE *out;
    int bytes;                  /* Bytes in buffer */
    int outbytes;               /* Bytes already written to file */
    int bofaoffs;               /* Current address relative to BOFA */
    uint16_t lineno;
    uint8_t *stringptr;        /* Accumulating string data, pointer to len */
    uint8_t data[253];
};

/* Finalize and output a data block */
static void writeblock(struct basic_block *blk, bool final)
{
    int bytes = blk->bytes;

    if (!bytes && !final)
        return;

    memset(blk->data + bytes + 1, 0, 252 - blk->bytes);
    blk->data[blk->bytes] = final ? 1 : 0;
    fwrite(blk->data, 1, 253, blk->out);

    blk->bofaoffs += blk->bytes;
    blk->outbytes += 253;
    blk->bytes = 0;
    blk->stringptr = NULL;
}

/*
 * Close a BASIC string being emitted.
 */
static void closestring(struct basic_block *blk)
{
    uint8_t *p = blk->data + blk->bytes;

    if (!blk->stringptr)
        return;

    *p++ = 0xbb;                /* Drop expression */
    *p++ = 0x0d;                /* Terminating CR */

    blk->bytes += 2;
    blk->stringptr = NULL;
}

/*
 * Emit data encapsulated in BASIC strings.  Multiple chunks can be
 * merged into common strings and strings can be broken up between
 * blocks.
 */
static void stringdata(struct basic_block *blk, const uint8_t * data, int len)
{
    while (len) {
        int copylen;
        uint8_t *p = blk->data + blk->bytes;

        if (!blk->stringptr) {
            if (blk->bytes >= 253 - 1 - 8) {
                /* Not enough space for any string */
                writeblock(blk, false);
                continue;
            }
            blk->stringptr = p;
            *p++ = 8;           /* Overhead per statement */
            p = set16(p, ++blk->lineno);
            *p++ = 0xcb;        /* String */
            *p++ = '"';         /* Which kind of quotation mark */
            *p++ = 0;           /* String length */
            blk->bytes += 6;
        }

        copylen = (253 - 1 - 2) - blk->bytes;
        if (copylen > len)
            copylen = len;

        if (copylen <= 0) {
            /* Not enough space left in this block for even one byte */
            closestring(blk);
            continue;
        }

        memcpy(p, data, copylen);

        p = blk->stringptr;
        p[0] += copylen;
        p[5] += copylen;

        blk->bytes += copylen;
        len -= copylen;
    }
}

/*
 * Emit a standalone BASIC statement, without encapsulating it in a string.
 */
static void bacstmt(struct basic_block *blk, const uint8_t * data, int len)
{
    uint8_t *p;

    if (blk->stringptr)
	closestring(blk);	/* Close any open string data */

    if (len > 253 - 1 - 4 - blk->bytes)
        writeblock(blk, false);

    p = &blk->data[blk->bytes];
    *p++ = len + 4;             /* Length of statement */
    p = set16(p, ++blk->lineno);        /* Line number */
    p = mempcpy(p, data, len);  /* Actual binary data */
    *p++ = 0x0d;                /* Terminating CR */

    blk->bytes = p - blk->data;
}

/*
 * BASIC statement for the original RUN
 *
 * PRINT CALL(PEEK(65054%)+SWAP%(65055%)-xxxx%)
 *
 * 65054 = EOFA, since we have the loader at the end
 */
static uint8_t run_prefix[] = {
    0x84, 0xc7, 0x1e, 0xfe, 0xce, 0x36, 0xc7, 0x1f,
    0xfe, 0xce, 0x36, 0xce, 0x34, 0xf5, 0xc7,
    0, 0,                       /* Loader length + 2 */
    0xf8, 0xce, 0x3a, 0xba
};
#define RUN_CALL_OFFSET 15

/*
 * This is the BASIC statement 1 POKE 65080%,CALL(yyyy%)
 * followed by an end of program marker (1 byte).  65080% is
 * the location of ERRCODE, which will be reset to -1 when
 * END is executed anyway, which will happen immediately.
 *
 * This includes the BASIC statement encapsulation, since this is
 * part of the payload.
 */
static uint8_t restart_prefix[] = {
    0x11, 0x01, 0x00, 0x86, 0x81, 0xc7,
    0x38, 0xfe,			/* Point to scratch byte (ERRCODE) */
    0xba, 0x2c, 0xc7,
    0, 0,			/* CALL address goes here */
    0xce, 0x3a, 0xba, 0x0d,
    0x01			/* End of BASIC program */
};
#define RESTART_CALL_OFFSET 11

/*
 * STOP statement, used as end-of-data marker.  Technically,
 * any BASIC statement of 3 bytes or less would work here.
 */
static const uint8_t stop_stmt[] = {
    0x86, 0x89			/* BASIC bytecode for STOP */
};

/*
 * Main routine
 */
int writebac80(FILE * out, const uint8_t * data, int len, int origin,
               int entrypoint, bool driver)
{
    struct basic_block blk;
    uint8_t *loader, *relocdata;
    int reloclen = 0;
    int loader_len;
    int loader_offset = 1;      /* 1 for the statement terminal CR */
    int pfxlen = 0;

    memset(&blk, 0, sizeof blk);
    blk.data[0] = 0x82;         /* Program start marker, not in memory */
    blk.bytes = 1;
    blk.out = out;

    if (have_relocs) {
        /* This is a relocatable binary, load it at BOFA */
        loader = bac80r_bin;
        loader_len = bac80r_SYM__end;
        loader_offset += (loader_len - bac80r_SYM__start);

        if (!driver) {
            /* Application, add a restart prefix to payload */
            pfxlen = sizeof restart_prefix;
            set16(restart_prefix + RESTART_CALL_OFFSET, entrypoint);
            push_reloc(RESTART_CALL_OFFSET - pfxlen);
        }
        loader[bac80r_SYM_DELTA] = pfxlen;
        set16(loader + bac80r_SYM_ENTRY, entrypoint + pfxlen);

        reloclen = pack_relocs(&relocdata, pfxlen, false);
        if (reloclen < 0)
            return -1;

        set16(loader + bac80r_SYM_NEG_RELOC_LEN, -(reloclen+1));
    } else {
        /* This is to be loaded at a fixed address */
        loader = bac80_bin;
        loader_len = bac80_SYM__end;
        loader_offset += (loader_len - bac80_SYM__start);

        set16(loader + bac80_SYM_ENTRY, entrypoint);
        set16(loader + bac80_SYM_LOAD, origin);
    }

    set16(run_prefix + RUN_CALL_OFFSET, loader_offset);

    /* Emit the RUN call prefix */
    bacstmt(&blk, run_prefix, sizeof run_prefix);

    /* Emit the restart prefix, if applicable */
    stringdata(&blk, restart_prefix, pfxlen);

    /* Emit the actual binary data */
    stringdata(&blk, data, len);

    /* Emit relocation information */
    if (have_relocs) {
        stringdata(&blk, (const uint8_t *)"\1", 1);     /* EOFA marker */
        stringdata(&blk, relocdata, reloclen);
    }

    /* Emit STOP statement to indicate end of data blocks */
    bacstmt(&blk, stop_stmt, sizeof stop_stmt);

    /* Emit loader */
    bacstmt(&blk, loader, loader_len);

    /* Write the final block */
    closestring(&blk);
    writeblock(&blk, true);

    return 0;
}
